package pps.core.space.service;

import cn.hutool.core.collection.CollUtil;
import cn.hutool.core.date.DateField;
import cn.hutool.core.date.DateTime;
import cn.hutool.core.date.DateUtil;
import cn.hutool.core.text.CharSequenceUtil;
import cn.hutool.core.util.ObjectUtil;
import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.github.pagehelper.PageInfo;
import com.github.pagehelper.page.PageMethod;
import pps.cloud.prediction.service.IThirdPowerCloudService;
import pps.cloud.prediction.service.data.third_active_power.GetThirdActivePowerInput;
import pps.cloud.prediction.service.data.third_active_power.GetThirdActivePowerOutput;
import pps.cloud.prediction.service.data.third_daily_accumulation_update.GetThirdDailyAccumulationUpdateInput;
import pps.cloud.prediction.service.data.third_daily_accumulation_update.GetThirdDailyAccumulationUpdateOutput;
import pps.cloud.prediction.service.data.wellhead_daily_production_situation.GetWellheadDailyProductionSituationInput;
import pps.cloud.prediction.service.data.wellhead_daily_production_situation.GetWellheadDailyProductionSituationOutput;
import pps.cloud.space.service.IDailyElectricityTrendCloudService;
import pps.cloud.space.service.data.line_daily_electricity_trend.GetLineDailyElectricityTrendInput;
import pps.cloud.space.service.data.line_daily_electricity_trend.GetLineDailyElectricityTrendOutput;
import pps.cloud.space.service.data.well_daily_electricity_trend.GetWellDailyElectricityTrendInput;
import pps.cloud.space.service.data.well_daily_electricity_trend.GetWellDailyElectricityTrendOutput;
import pps.core.common.constant.BusinessConstant;
import pps.core.common.utils.BaseUtils;
import pps.core.space.entity.LineDailyElectricityTrendEnt;
import pps.core.space.entity.LineDailyElectricityTrendView;
import pps.core.space.entity.SpaceOptimizeShortDurationView;
import pps.core.space.entity.WellDailyElectricityTrendView;
import pps.core.space.mapper.LineDailyElectricityTrendMapper;
import pps.core.space.mapper.LineDailyElectricityTrendViewMapper;
import pps.core.space.mapper.SpaceOptimizeShortDurationViewMapper;
import pps.core.space.mapper.WellDailyElectricityTrendViewMapper;
import xstartup.annotation.XService;
import xstartup.base.XContext;
import xstartup.base.util.XCopyUtils;
import xstartup.data.XListResult;
import xstartup.data.XPageResult;
import xstartup.data.XServiceResult;
import xstartup.helper.XTransactionHelper;

import java.math.BigDecimal;
import java.util.*;
import java.util.function.Function;
import java.util.stream.Collectors;

/**
 * 日用电趋势计算Cloud模块
 *
 * @author ZWT
 * @date 2024/05/22 17:17
 */
@XService
public class DailyElectricityTrendCloudServiceImpl implements IDailyElectricityTrendCloudService {

    /**
     * 日用电趋势计算
     * <p>
     * (1)取昨日极短期间开优化结果
     * (2)绿电阶段开井时间：每15分钟判段光伏出力>运行功率（考虑开井顺序,功率累加） 并且当前时段开井，则算坐开井时段(0.25小时)，全天累加
     * (3)单井绿电用电：运行功率*绿电阶段开井时间
     * (4)市电用电:总用电-光伏用电
     *
     * @param context 上下文
     * @return {@link XServiceResult }
     */
    @Override
    public XServiceResult dailyElectricityTrendJob(XContext context) {
        DateTime yesterday = DateUtil.beginOfDay(DateUtil.yesterday());
        //查每15分钟的有功功率
        Map<String, List<GetThirdActivePowerOutput>> powerMap = this.getAvgPhotovoltaicPower(context, yesterday);
        //查昨日间开优化结果
        Map<String, List<SpaceOptimizeShortDurationView>> lineMap = this.getOptimizationResult(context, yesterday);
        //查昨日井口生产情况
        Map<String, GetWellheadDailyProductionSituationOutput> wellProductionMap = this.getWellProductionList(context, yesterday);
        //查昨日日累计用电
        Map<String, BigDecimal> dailyUpdatePowerMap = this.getDailyUpdatePowerMap(context, yesterday);
        //查去年同期用电趋势
        Map<String, LineDailyElectricityTrendEnt> lastElectricityTrendMap = this.getLastElectricityTrendMap(context, yesterday);
        //用电趋势计算结果
        List<LineDailyElectricityTrendView> stationSaveList = new ArrayList<>(powerMap.size());
        //井口用电趋势计算结果
        List<WellDailyElectricityTrendView> wellSaveList = new ArrayList<>(wellProductionMap.size());
        SpaceOptimizeShortDurationView wellInfo;
        BigDecimal currentServiceRating;
        String wellNumber;
        List<GetThirdActivePowerOutput> thirdActivePowerList;
        Map<Integer, List<SpaceOptimizeShortDurationView>> startSeqMap;
        List<SpaceOptimizeShortDurationView> durationList;
        GetThirdActivePowerOutput thirdActivePower;
        BigDecimal dailyElectricityConsumption;
        //遍历全量电站
        for (String stationName : powerMap.keySet()) {
            //取存在的间开
            if (lineMap.containsKey(stationName)) {
                //每15分钟有功功率
                thirdActivePowerList = powerMap.get(stationName);
                //按照启动顺序分组
                startSeqMap = lineMap.get(stationName).stream()
                        .collect(Collectors.groupingBy(SpaceOptimizeShortDurationView::getStartSeq));
                //光伏日累计用电(绿电)
                BigDecimal photovoltaicPower = BigDecimal.ZERO;
                //按启动顺序取时间段列表
                BigDecimal serviceRating = BigDecimal.ZERO;
                for (int i = 1; i <= startSeqMap.size(); i++) {
                    durationList = startSeqMap.get(i);
                    //当前井口
                    wellInfo = durationList.get(0);
                    //当前井运行功率
                    currentServiceRating = wellInfo.getServiceRating();
                    //井号
                    wellNumber = wellInfo.getWellNumber();
                    //累加运行功率
                    serviceRating.add(currentServiceRating);
                    //绿电开井时间(h)
                    BigDecimal dailyGreenOpenHour = BigDecimal.ZERO;
                    //记录每一次计算结束位置
                    int index = 0;
                    for (SpaceOptimizeShortDurationView duration : durationList) {
                        //时间转换
                        DateTime openWellTime = DateUtil.parse(duration.getOpenWellTime() + BusinessConstant.INITIALIZATION_SECOND, BusinessConstant.TIME_FORMAT);
                        DateTime closeWellTime;
                        if (CharSequenceUtil.equals(duration.getCloseWellTime(), BusinessConstant.END_OF_DAY_TIME)) {
                            closeWellTime = BusinessConstant.DATE_FLAG;
                        } else {
                            closeWellTime = DateUtil.parse(duration.getCloseWellTime() + BusinessConstant.INITIALIZATION_SECOND, BusinessConstant.TIME_FORMAT);
                        }
                        //遍历功率
                        for (int i1 = index; i1 < thirdActivePowerList.size(); i1++) {
                            thirdActivePower = thirdActivePowerList.get(i1);
                            //在时间范围内,并且总运行功率满足运行
                            if (DateUtil.isIn(thirdActivePower.getCreateDate(), openWellTime, closeWellTime) && thirdActivePower.getPhotovoltaicPower().compareTo(serviceRating) >= 0) {
                                //当前井口绿电运行时长(累加0.25小时)
                                dailyGreenOpenHour = dailyGreenOpenHour.add(BusinessConstant.BIG_DECIMAL_0_25);
                            }
                        }
                    }
                    //单井绿电用电：运行功率*绿电阶段开井时间
                    BigDecimal dailyGreenConsumption = dailyGreenOpenHour.multiply(currentServiceRating);
                    //井场绿电用电累加
                    photovoltaicPower = photovoltaicPower.add(dailyGreenConsumption);
                    //用电趋势结果计算
                    wellSaveList.add(this.getWellDailyElectricityTrendView(wellProductionMap.get(wellNumber), wellNumber,
                            wellInfo.getRunDuration(), dailyGreenOpenHour, dailyGreenConsumption, yesterday));
                }
                //计算封装井场用电趋势数据
                dailyElectricityConsumption = Optional.ofNullable(dailyUpdatePowerMap.get(stationName)).orElse(BigDecimal.ZERO);
                stationSaveList.add(
                        this.getLineDailyElectricityTrendView(lastElectricityTrendMap.get(stationName),
                                dailyElectricityConsumption, photovoltaicPower, stationName, yesterday)
                );
            }
        }
        //开启事务批量插入
        return XTransactionHelper.begin(context, () -> {
            int size;
            if (CollUtil.isNotEmpty(stationSaveList)) {
                LineDailyElectricityTrendViewMapper mapper = context.getBean(LineDailyElectricityTrendViewMapper.class);
                size = stationSaveList.size();
                if (size > BaseUtils.BATCH_SIZE) {
                    List<List<LineDailyElectricityTrendView>> subList = BaseUtils.getSubList(stationSaveList);
                    subList.forEach(mapper::batchInsert);
                } else {
                    mapper.batchInsert(stationSaveList);
                }
            }
            if (CollUtil.isNotEmpty(wellSaveList)) {
                WellDailyElectricityTrendViewMapper mapper = context.getBean(WellDailyElectricityTrendViewMapper.class);
                size = wellSaveList.size();
                if (size > BaseUtils.BATCH_SIZE) {
                    List<List<WellDailyElectricityTrendView>> subList = BaseUtils.getSubList(wellSaveList);
                    subList.forEach(mapper::batchInsert);
                } else {
                    mapper.batchInsert(wellSaveList);
                }
            }
            return XServiceResult.OK;
        });
    }

    /**
     * 用电趋势统计
     *
     * @param context 上下文
     * @param input   输入
     * @return {@link XListResult }<{@link GetLineDailyElectricityTrendOutput }>
     */
    @Override
    public XListResult<GetLineDailyElectricityTrendOutput> queryDailyElectricityTrend(XContext context, GetLineDailyElectricityTrendInput input) {
        LineDailyElectricityTrendViewMapper mapper = context.getBean(LineDailyElectricityTrendViewMapper.class);
        List<LineDailyElectricityTrendView> list = mapper.selectDailyElectricityTrend(
                LineDailyElectricityTrendView.builder()
                        .startTime(input.getStartTime())
                        .endTime(input.getEndTime())
                        .stationNameList(input.getStationNameList())
                        .build()
        );
        return XListResult.success(XCopyUtils.copyNewList(list, GetLineDailyElectricityTrendOutput.class));
    }

    /**
     * 井口分析分页
     *
     * @param context 上下文
     * @param input   输入
     * @return {@link XPageResult }<{@link GetWellDailyElectricityTrendOutput }>
     */
    @Override
    public XPageResult<GetWellDailyElectricityTrendOutput> queryWellheadAnalysisPage(XContext context, GetWellDailyElectricityTrendInput input) {
        DateTime date = DateUtil.date(input.getCreateDate());
        PageMethod.startPage(input.getPage(), input.getLimit());
        WellDailyElectricityTrendViewMapper mapper = context.getBean(WellDailyElectricityTrendViewMapper.class);
        List<WellDailyElectricityTrendView> list = mapper.selectWellheadAnalysisPage(
                WellDailyElectricityTrendView.builder()
                        .createDate(date)
                        .month(date.monthBaseOne())
                        .year(date.year())
                        .wellNumberList(input.getWellNumberList())
                        .wellNumber(input.getWellNumber())
                        .build()
        );
        PageInfo<WellDailyElectricityTrendView> pageInfo = new PageInfo<>(list);
        List<GetWellDailyElectricityTrendOutput> output;
        if (CollUtil.isNotEmpty(list)) {
            output = new ArrayList<>(list.size());
            for (WellDailyElectricityTrendView view : list) {
                output.add(GetWellDailyElectricityTrendOutput.builder()
                        .wellNumber(view.getWellNumber())
                        .spaceRunDuration(view.getSpaceRunDuration())
                        .dailyLiquidProduction(view.getDailyLiquidProduction())
                        .greenElectricityRate(view.getGreenElectricityRate())
                        .dailyCarbonReduction(view.getDailyCarbonReduction())
                        .monthGreenElectricityRate(BaseUtils.getRate(view.getMonthDailyGreenOpenHour(), view.getMonthSpaceRunDuration()))
                        .monthDailyCarbonReduction(view.getMonthDailyGreenConsumption().multiply(BusinessConstant.BIG_DECIMAL_6_67))
                        .yearGreenElectricityRate(BaseUtils.getRate(view.getYearDailyGreenOpenHour(), view.getYearSpaceRunDuration()))
                        .yearDailyCarbonReduction(view.getYearDailyGreenConsumption().multiply(BusinessConstant.BIG_DECIMAL_6_67))
                        .build());
            }
        } else {
            output = Collections.emptyList();
        }
        return XPageResult.success(output, input, pageInfo.getTotal());
    }

    /*------------------------------------- private -------------------------------------*/

    /**
     * 获取同期用电趋势信息
     *
     * @param context    上下文
     * @param createDate 创建日期
     * @return {@link Map }<{@link String }, {@link LineDailyElectricityTrendEnt }>
     */
    private Map<String, LineDailyElectricityTrendEnt> getLastElectricityTrendMap(XContext context, Date createDate) {
        LineDailyElectricityTrendMapper mapper = context.getBean(LineDailyElectricityTrendMapper.class);
        List<LineDailyElectricityTrendEnt> list = mapper.selectList(new LambdaQueryWrapper<LineDailyElectricityTrendEnt>()
                .apply("DATE( create_date ) = {0}", DateUtil.offset(createDate, DateField.YEAR, -1))
        );
        Map<String, LineDailyElectricityTrendEnt> collect;
        if (CollUtil.isEmpty(list)) {
            collect = Collections.emptyMap();
        } else {
            collect = list.stream()
                    .collect(Collectors.toMap(LineDailyElectricityTrendEnt::getStationName, Function.identity()));
        }
        return collect;
    }

    /**
     * 查询每日井场日累计数据
     *
     * @param context    上下文
     * @param createDate 创建日期
     * @return {@link Map }<{@link String }, {@link BigDecimal }>
     */
    private Map<String, BigDecimal> getDailyUpdatePowerMap(XContext context, Date createDate) {
        IThirdPowerCloudService service = context.getBean(IThirdPowerCloudService.class);
        XListResult<GetThirdDailyAccumulationUpdateOutput> result = service.queryDailyUpdatePower(context, GetThirdDailyAccumulationUpdateInput.builder()
                .createDate(createDate)
                .build());
        result.throwIfFail();
        Map<String, BigDecimal> powerMap;
        List<GetThirdDailyAccumulationUpdateOutput> powerList = result.getResult();
        if (CollUtil.isEmpty(powerList)) {
            powerMap = Collections.emptyMap();
        } else {
            powerMap = powerList.stream()
                    .collect(Collectors.toMap(GetThirdDailyAccumulationUpdateOutput::getStationName, GetThirdDailyAccumulationUpdateOutput::getDailyElectricityConsumption));
        }
        return powerMap;
    }

    /**
     * 查每15分钟的有功功率
     *
     * @param context    上下文
     * @param createDate 创建日期
     * @return {@link Map }<{@link String }, {@link List }<{@link GetThirdActivePowerOutput }>>
     */
    private Map<String, List<GetThirdActivePowerOutput>> getAvgPhotovoltaicPower(XContext context, Date createDate) {
        IThirdPowerCloudService service = context.getBean(IThirdPowerCloudService.class);
        XListResult<GetThirdActivePowerOutput> result = service.queryAvgPhotovoltaicPower(context, GetThirdActivePowerInput.builder()
                .createDate(createDate)
                .build());
        result.throwIfFail();
        Map<String, List<GetThirdActivePowerOutput>> powerMap;
        List<GetThirdActivePowerOutput> powerList = result.getResult();
        if (CollUtil.isEmpty(powerList)) {
            powerMap = Collections.emptyMap();
        } else {
            powerMap = powerList.stream()
                    .collect(Collectors.groupingBy(GetThirdActivePowerOutput::getStationName));
        }
        return powerMap;
    }

    /**
     * 每日井口生产情况
     *
     * @param context    上下文
     * @param createDate 创建日期
     * @return {@link Map }<{@link String }, {@link GetWellheadDailyProductionSituationOutput }>
     */
    private Map<String, GetWellheadDailyProductionSituationOutput> getWellProductionList(XContext context, Date createDate) {
        IThirdPowerCloudService service = context.getBean(IThirdPowerCloudService.class);
        XListResult<GetWellheadDailyProductionSituationOutput> result = service.queryWellProductionList(context, GetWellheadDailyProductionSituationInput.builder()
                .createDate(createDate)
                .build());
        result.throwIfFail();
        Map<String, GetWellheadDailyProductionSituationOutput> collect;
        List<GetWellheadDailyProductionSituationOutput> list = result.getResult();
        if (CollUtil.isEmpty(list)) {
            collect = Collections.emptyMap();
        } else {
            collect = list.stream()
                    .collect(Collectors.toMap(GetWellheadDailyProductionSituationOutput::getWellNumber, Function.identity()));
        }
        return collect;
    }

    /**
     * 获取优化结果
     *
     * @param context      上下文
     * @param optimizeDate 优化日期
     * @return {@link Map }<{@link String }, {@link List }<{@link SpaceOptimizeShortDurationView }>>
     */
    private Map<String, List<SpaceOptimizeShortDurationView>> getOptimizationResult(XContext context, Date optimizeDate) {
        SpaceOptimizeShortDurationViewMapper shortMapper = context.getBean(SpaceOptimizeShortDurationViewMapper.class);
        List<SpaceOptimizeShortDurationView> durationList = shortMapper.selectOptimizationResultList(SpaceOptimizeShortDurationView.builder()
                .optimizeDate(optimizeDate)
                .build());
        Map<String, List<SpaceOptimizeShortDurationView>> lineMap;
        if (CollUtil.isEmpty(durationList)) {
            lineMap = Collections.emptyMap();
        } else {
            lineMap = durationList.stream()
                    .collect(Collectors.groupingBy(SpaceOptimizeShortDurationView::getLineName));
        }
        return lineMap;
    }

    /**
     * 获取井口日用电趋势信息
     *
     * @param productionSituationOutput 生产情况输出
     * @param wellNumber                井号
     * @param runDuration               运行持续时间
     * @param dailyGreenOpenHour        每日绿色开放时间
     * @param dailyGreenConsumption     绿电用电
     * @param yesterday                 昨天
     * @return {@link WellDailyElectricityTrendView }
     */
    private WellDailyElectricityTrendView getWellDailyElectricityTrendView(GetWellheadDailyProductionSituationOutput productionSituationOutput,
                                                                           String wellNumber, BigDecimal runDuration, BigDecimal dailyGreenOpenHour,
                                                                           BigDecimal dailyGreenConsumption, Date yesterday) {
        WellDailyElectricityTrendView view;
        if (ObjectUtil.isNotNull(productionSituationOutput)) {
            view = WellDailyElectricityTrendView.builder()
                    .wellNumber(wellNumber)
                    .spaceRunDuration(runDuration)
                    .dailyOpenHour(productionSituationOutput.getProductionTime())
                    .dailyGreenOpenHour(dailyGreenOpenHour)
                    .dailyGreenConsumption(dailyGreenConsumption)
                    .dailyElectricityConsumption(productionSituationOutput.getDailyElectricityConsumption())
                    .dailyLiquidProduction(productionSituationOutput.getDailyLiquidProduction())
                    //日减碳量:日绿电用电量*6.67
                    .dailyCarbonReduction(dailyGreenConsumption.multiply(BusinessConstant.BIG_DECIMAL_6_67))
                    //绿电占比:绿电开井时长/开井时长*100%
                    .greenElectricityRate(BaseUtils.getRate(dailyGreenOpenHour, runDuration))
                    .createDate(yesterday)
                    .systemSource("长庆")
                    .build();
        } else {
            view = WellDailyElectricityTrendView.builder()
                    .wellNumber(wellNumber)
                    .spaceRunDuration(runDuration)
                    .dailyOpenHour(BigDecimal.ZERO)
                    .dailyGreenOpenHour(dailyGreenOpenHour)
                    .dailyGreenConsumption(dailyGreenConsumption)
                    .dailyElectricityConsumption(BigDecimal.ZERO)
                    .dailyLiquidProduction(BigDecimal.ZERO)
                    //日减碳量:日绿电用电量*6.67
                    .dailyCarbonReduction(dailyGreenConsumption.multiply(BusinessConstant.BIG_DECIMAL_6_67))
                    //绿电占比:绿电开井时长/开井时长*100%
                    .greenElectricityRate(BaseUtils.getRate(dailyGreenOpenHour, runDuration))
                    .createDate(yesterday)
                    .systemSource("长庆")
                    .build();
        }
        return view;
    }

    /**
     * 获取线路日电量趋势
     *
     * @param ent                         ent
     * @param dailyElectricityConsumption 日用电量
     * @param photovoltaicPower           光伏发电
     * @param stationName                 车站名称
     * @param yesterday                   昨天
     * @return {@link LineDailyElectricityTrendView }
     */
    private LineDailyElectricityTrendView getLineDailyElectricityTrendView(LineDailyElectricityTrendEnt ent, BigDecimal dailyElectricityConsumption,
                                                                           BigDecimal photovoltaicPower, String stationName, Date yesterday) {
        LineDailyElectricityTrendView view;
        BigDecimal municipalPower = dailyElectricityConsumption.subtract(photovoltaicPower);
        if (municipalPower.compareTo(BigDecimal.ZERO) < 0) {
            municipalPower = BigDecimal.ZERO;
        }
        if (ObjectUtil.isNotNull(ent)) {
            view = LineDailyElectricityTrendView.builder()
                    .stationName(stationName)
                    .totalPower(dailyElectricityConsumption)
                    .photovoltaicPower(photovoltaicPower)
                    //市电日用电量:总用电-光伏用电
                    .municipalPower(municipalPower)
                    .lastTotalPower(ent.getTotalPower())
                    .lastPhotovoltaicPower(ent.getPhotovoltaicPower())
                    .lastMunicipalPower(ent.getMunicipalPower())
                    //同比:本期/上期*100
                    .totalRatio(BaseUtils.getRate(dailyElectricityConsumption, ent.getTotalPower()))
                    .photovoltaicRatio(BaseUtils.getRate(photovoltaicPower, ent.getPhotovoltaicPower()))
                    .municipalRatio(BaseUtils.getRate(municipalPower, ent.getMunicipalPower()))
                    .createDate(yesterday)
                    .systemSource("长庆")
                    .build();
        } else {
            view = LineDailyElectricityTrendView.builder()
                    .stationName(stationName)
                    .totalPower(dailyElectricityConsumption)
                    .photovoltaicPower(photovoltaicPower)
                    //市电日用电量:总用电-光伏用电
                    .municipalPower(municipalPower)
                    .lastTotalPower(BigDecimal.ZERO)
                    .lastPhotovoltaicPower(BigDecimal.ZERO)
                    .lastMunicipalPower(BigDecimal.ZERO)
                    .totalRatio(BigDecimal.ZERO)
                    .photovoltaicRatio(BigDecimal.ZERO)
                    .municipalRatio(BigDecimal.ZERO)
                    .createDate(yesterday)
                    .systemSource("长庆")
                    .build();
        }
        return view;
    }
}